/***********************************************************************************************************************
 *
 * blueBill Core - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it
 * SCM: https://kenai.com/hg/bluebill~core-src
 *
 **********************************************************************************************************************/
package it.tidalwave.mobile.media;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import javax.annotation.concurrent.Immutable;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.openide.util.Lookup;
import it.tidalwave.util.As;
import it.tidalwave.util.AsException;
import it.tidalwave.util.Id;
import it.tidalwave.util.Key;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.util.TypeSafeMap;
import it.tidalwave.util.TypeSafeHashMap;
import it.tidalwave.role.spi.DefaultIdentifiable;
import it.tidalwave.role.Identifiable;
import it.tidalwave.role.Removable;
import it.tidalwave.mobile.util.Downloadable;
import it.tidalwave.mobile.util.DefaultDownloadable;
import it.tidalwave.netbeans.util.AsLookupSupport;
import it.tidalwave.semantic.GenericEntity;
import it.tidalwave.semantic.vocabulary.DublinCoreVocabulary;

/***********************************************************************************************************************
 *
 * FIXME: refactor over EntityWithProperties! 
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@Immutable
public class Media extends AsLookupSupport implements As, Lookup.Provider
  {
    public static final Class<Media> Media = Media.class;
    
    private final static long serialVersionUID = 876756656347568673L;
    
    private static final Logger log = LoggerFactory.getLogger(Media.class);

    public static final Key<GenericEntity> AUTHOR = new Key<GenericEntity>(DublinCoreVocabulary.ID_DC_CREATOR);
    public static final Key<GenericEntity> SOURCE = new Key<GenericEntity>(DublinCoreVocabulary.ID_DC_PUBLISHER);
    public static final Key<String> TITLE = new Key<String>(DublinCoreVocabulary.ID_DC_TITLE);
    public static final Key<Date> DATE = new Key<Date>(DublinCoreVocabulary.ID_DC_DATE);
    public static final Key<Id> ID = new Key<Id>(DublinCoreVocabulary.ID_DC_IDENTIFIER);
    public static final Key<String> RIGHTS = new Key<String>(DublinCoreVocabulary.ID_DC_RIGHTS);
    public static final Key<String> FORMAT = new Key<String>(DublinCoreVocabulary.ID_DC_FORMAT);
    public static final Key<String> TYPE = new Key<String>(DublinCoreVocabulary.ID_DC_TYPE);

    // FIXME: remove, use standard vocabularies.
    public static final String XX = "urn:bluebill-mobile:";
    public static final Key<URI> RES_URL = new Key<URI>(XX + "url");
    public static final Key<String> MIME_TYPE = new Key<String>(XX + "mime-type");
    public static final Key<Integer> DURATION = new Key<Integer>(XX + "duration");

    @Nonnull
    private final TypeSafeMap propertyMap;

    @CheckForNull
    private Downloadable downloadable;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public Media()
      {
        this(Collections.<Key<?>, Object>emptyMap());
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private Media (final @Nonnull Map<Key<?>, Object> propertyMap)
      {
        this.propertyMap = new TypeSafeHashMap(propertyMap);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public <T> T get (final @Nonnull Key<T> key)
      throws NotFoundException
      {
        return propertyMap.get(key);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public boolean contains (final @Nonnull Key<?> item)
      {
        return propertyMap.containsKey(item);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public Set<Key<?>> getKeys()
      {
        return propertyMap.getKeys();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnegative
    public int getSize()
      {
        return propertyMap.getSize();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public Iterator<Object> iterator()
      {
        return propertyMap.iterator();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public <T> Media with (final @Nonnull Key<T> key, final @Nonnull T value)
      {
        if (key.equals(FORMAT))
          {
            final String[] parts = ((String)value).split(";");
            final String mimeType = parts[0];
            Media media = with(MIME_TYPE, mimeType);

            if (parts.length > 1)
              {
                final String durationAsString = parts[1].replaceAll("s$", "").trim();

                if (!durationAsString.equals(""))
                  {
                    media = media.with(DURATION, Integer.valueOf(durationAsString));
                  }
              }

            return media;
          }

        final Map<Key<?>, Object> propertyMapAsPlainMap = propertyMap.asMap();
        propertyMapAsPlainMap.put(key, value);

        propertyMapAsPlainMap.remove(FORMAT);

        if (propertyMapAsPlainMap.containsKey(MIME_TYPE))
          {
            propertyMapAsPlainMap.remove(TYPE);

            if (((String)propertyMapAsPlainMap.get(MIME_TYPE)).startsWith("image/"))
              {
                propertyMapAsPlainMap.put(TYPE, DublinCoreVocabulary.TYPE_STILL_IMAGE.stringValue());
              }
            else if(((String) propertyMapAsPlainMap.get(MIME_TYPE)).startsWith("audio/"))
              {
                propertyMapAsPlainMap.put(TYPE, DublinCoreVocabulary.TYPE_SOUND.stringValue());
              }
          }

        if (propertyMapAsPlainMap.containsKey(MIME_TYPE) || propertyMapAsPlainMap.containsKey(DURATION))
          {
            final StringBuilder buffer = new StringBuilder();
            final Key<?>[] keys = new Key<?>[] { MIME_TYPE, DURATION };

            String separator = "";

            for (final Key<?> k : keys)
              {
                buffer.append(separator);

                if (k.equals(MIME_TYPE))
                  {
                    buffer.append(propertyMapAsPlainMap.get(k));
                    separator = "; ";
                  }
                else if (propertyMapAsPlainMap.containsKey(k))
                  {
                    buffer.append(propertyMapAsPlainMap.get(k)).append("s");
                    separator = "; ";
                  }
              }

            if (buffer.length() > 0)
              {
                propertyMapAsPlainMap.put(FORMAT, buffer.toString());
              }
          }

        return new Media(propertyMapAsPlainMap);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public <T> T as (final @Nonnull Class<T> type, final @Nonnull NotFoundBehaviour<T> nfb)
      {
        if (type.equals(Identifiable.class))
          {
            try
              {
                return type.cast(new DefaultIdentifiable(get(ID)));
              }
            catch (NotFoundException e)
              {
                throw new AsException(type, e);
              }
          }
        
        if (type.equals(Downloadable.class) || type.equals(Removable.class))
          {
            return super.as(type, new NotFoundBehaviour<T>()
              {
                @Nonnull
                public synchronized T run (final Throwable t)
                  {
                    if (downloadable == null)
                      {
                        try
                          {
                            downloadable = new DefaultDownloadable(new URL(Media.this.get(ID).stringValue()));
                          }
                        catch (MalformedURLException e)
                          {
                            return nfb.run(new AsException(type, e));
                          }
                        catch (NotFoundException e)
                          {
                            return nfb.run(new AsException(type, e));
                          }
                      }

                    return type.cast(downloadable);
                  }
              });
          }

        return super.as(type, nfb);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public String toString()
      {
        try
          {
            return String.format("Media@%x[%s, %s]", System.identityHashCode(this), propertyMap.get(ID), propertyMap);
          }
        catch (NotFoundException e)
          {
            return String.format("Media@%x[%s]", System.identityHashCode(this), propertyMap);
          }
      }
  }
